package gs.test.v0.api.trace

import cats.~>
import cats.Applicative
import cats.effect.kernel.Async
import cats.effect.kernel.Ref
import cats.effect.kernel.Resource
import cats.effect.syntax.all.*
import cats.syntax.all.*
import java.net.URI
import natchez.*

/** Essentially copied from the "IOTrace" implementation within the excellent
  * Natchez, but updated to work with `Ref`.
  */
final class GsTestTrace[F[_]: Async] private (
  src: Ref[F, Span[F]]
) extends natchez.Trace[F]:

  def put(fields: (String, TraceValue)*): F[Unit] =
    src.get.flatMap(_.put(fields*))

  def log(fields: (String, TraceValue)*): F[Unit] =
    src.get.flatMap(_.log(fields*))

  def log(event: String): F[Unit] =
    src.get.flatMap(_.log(event))

  def attachError(
    err: Throwable,
    fields: (String, TraceValue)*
  ): F[Unit] =
    src.get.flatMap(_.attachError(err, fields*))

  def kernel: F[Kernel] = src.get.flatMap(_.kernel)

  def spanR(
    name: String,
    options: Span.Options = Span.Options.Defaults
  ): Resource[F, F ~> F] =
    for {
      parent <- Resource.eval(src.get)
      child  <- parent.span(name, options)
    } yield new (F ~> F) {
      def apply[A](fa: F[A]): F[A] =
        src.get.flatMap { old =>
          src
            .set(child)
            .bracket(_ => fa.onError { case e => child.attachError(e) })(_ =>
              src.set(old)
            )
        }

    }

  def span[A](
    name: String,
    options: Span.Options = Span.Options.Defaults
  )(
    k: F[A]
  ): F[A] =
    spanR(name, options).use(_(k))

  def traceId: F[Option[String]] =
    src.get.flatMap(_.traceId)

  override def spanId(
    implicit
    F: Applicative[F]
  ): F[Option[String]] =
    src.get.flatMap(_.spanId)

  override def traceUri: F[Option[URI]] = src.get.flatMap(_.traceUri)

object GsTestTrace:

  /** Initialize a new `natchez.Trace` instance based on a given Span.
    */
  def initialize[F[_]: Async](
    sourceSpan: Span[F]
  ): F[natchez.Trace[F]] =
    Ref.of[F, Span[F]](sourceSpan).map(src => new GsTestTrace[F](src))

end GsTestTrace
